import os
import signal
import subprocess
import time
from typing import Optional, List, Dict, Any

import httpx


class LocalTestServer:

    def __init__(self, *, model_uri: str,
                 registry_host: str,
                 registry_token: str,
                 test_serving_host: str = "0.0.0.0",
                 test_serving_port: int = 5000,
                 registry_is_uc: bool = False,
                 additional_serving_flags: Optional[List[str]] = None):
        self._model_uri = model_uri
        self._databricks_registry_host = registry_host
        self._databricks_registry_token = registry_token
        self._test_serving_host = test_serving_host
        self._test_serving_port = test_serving_port
        self._registry_is_uc = registry_is_uc
        self._additional_serving_flags = additional_serving_flags or []

        self._server_process = None
        self._http_client = httpx.Client(base_url=f"http://{self._test_serving_host}:{self._test_serving_port}")

    def start(self):
        subprocess.run(f"kill $(lsof -t -i:{self._test_serving_port})", shell=True)
        command_args = ["mlflow", "models", "serve", "-m", self._model_uri, "-p", str(self._test_serving_port),
                        *self._additional_serving_flags]
        # spawn in new process group
        current_env = os.environ.copy()
        current_env["DATABRICKS_HOST"] = self._databricks_registry_host
        current_env["DATABRICKS_TOKEN"] = self._databricks_registry_token
        self._server_process = subprocess.Popen(command_args,
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.PIPE,
                                                preexec_fn=os.setsid,
                                                env=current_env)

    def wait_and_assert_healthy(self):
        assert self._server_process is not None, "Server process has not been started."
        while True:
            if self._server_process.returncode is not None:
                stdout, stderr = self._server_process.communicate(timeout=10)
                print('STDOUT:', stdout.decode())
                print('STDERR:', stderr.decode())
                print('Exit Code:', self._server_process.returncode)
                raise ValueError("Server process has terminated unexpectedly.")

            try:
                resp = self._http_client.get("/health")
                if resp.status_code == 200:
                    print("Success")
                    break
            except Exception as e:
                print("endpoint not yet available")
            assert self._server_process.returncode is None, "Server process has terminated unexpectedly."
            time.sleep(1)

    def query(self, *, payload: Dict[str, Any]):
        return self._http_client.post("/invocations", json=payload)

    def stop(self):
        # Send SIGTERM to the subprocess
        os.killpg(os.getpgid(self._server_process.pid), signal.SIGTERM)

        # Wait for a short while and check if the process has terminated
        time.sleep(5)
        if self._server_process.poll() is None:
            # If still running, send SIGKILL
            os.killpg(os.getpgid(self._server_process.pid), signal.SIGKILL)

        # # Optionally, wait for the process to terminate and get its exit status
        try:
            # Wait for the process to terminate, with a timeout
            stdout, stderr = self._server_process.communicate(timeout=10)
        except subprocess.TimeoutExpired:
            stdout, stderr = self._server_process.communicate()
            print("Timed out")

        print('STDOUT:', stdout.decode())
        print('STDERR:', stderr.decode())
        print('Exit Code:', self._server_process.returncode)



import abc
import os
import random
import socket
import subprocess
import time
from dataclasses import field, dataclass
from typing import List, Optional

import httpx
from filelock import FileLock


def debug_msg(msg: str):
    print(f"[DEBUG][pid:{os.getpid()}] {msg}")


def is_port_open(host: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        result = sock.connect_ex((host, port))
        return result == 0


@dataclass(frozen=True, kw_only=True)
class EngineConfig(abc.ABC):
    model: str
    host: str = field(default="0.0.0.0")
    port: int = field(default=9989)
    openai_api_path: int = field(default="v1")

    @abc.abstractmethod
    def to_run_command(self, local_model_path: Optional[str] = None) -> List[str]:
        pass

    def default_pip_reqs(self, *, filelock_version: str = "3.15.4", **kwargs) -> List[str]:
        return [*self.engine_pip_reqs(**kwargs), f"filelock=={filelock_version}"]

    @abc.abstractmethod
    def engine_pip_reqs(self, **kwargs) -> List[str]:
        pass


class EngineProcess(abc.ABC):

    def __init__(self, *, config: EngineConfig):
        self._config = config
        self._lock = FileLock(f"{self.__class__.__name__}.txt.lock")
        self._server_http_client = httpx.Client(
            base_url=f"http://{self._config.host}:{self._config.port}",
            timeout=30
        )
        self._oai_http_client = httpx.Client(
            base_url=f"http://{self._config.host}:{self._config.port}/{self._config.openai_api_path}",
            timeout=30
        )
        self._proc = None

    @property
    def oai_http_client(self) -> httpx.Client:
        return self._oai_http_client

    @property
    def server_http_client(self) -> httpx.Client:
        return self._server_http_client

    @property
    @abc.abstractmethod
    def engine_name(self) -> str:
        pass

    @property
    def config(self) -> EngineConfig:
        return self._config

    @abc.abstractmethod
    def health_check(self) -> bool:
        pass

    def is_process_healthy(self) -> bool:
        if not is_port_open(self.config.host, self.config.port):
            return False

        return self.health_check()

    def stop_proc(self):
        with self._lock:
            if self.is_process_healthy() is True:
                subprocess.run(f"kill $(lsof -t -i:{self.config.port})", shell=True)

    # todo add local lora paths
    def start_proc(self, local_model_path: Optional[str] = None):
        # kill process in port if already running
        time.sleep(random.randint(1, 5))
        debug_msg(f"Attempting to acquire Lock")
        with self._lock:
            debug_msg(f"Acquired Lock")
            if self.health_check() is False:
                proc_env = {"HOST": self.config.host, "PORT": str(self.config.host)}
                self._proc = subprocess.Popen(self.config.to_run_command(local_model_path),
                                              env=proc_env)
                while self.is_process_healthy() is False:
                    debug_msg(f"Waiting for {self.engine_name} to start")
                    time.sleep(1)
            else:
                debug_msg(f"{self.engine_name} already running")



import json
from pathlib import Path
from typing import List, Dict, Type, Optional

import mlflow
from huggingface_hub import snapshot_download

from mlflow_extensions.serving.engines.base import EngineProcess, debug_msg, EngineConfig


class CustomServingEnginePyfuncWrapper(mlflow.pyfunc.PythonModel):

    # todo support lora modules
    def __init__(self,
                 *,
                 engine: Type[EngineProcess],
                 engine_config: EngineConfig,
                 model_artifact_key="model",
                 endpoint="/chat/completions"):
        self._engine_klass: Type[EngineProcess] = engine
        self._engine_config: EngineConfig = engine_config
        self._engine: Optional[EngineProcess] = None
        # for convenience
        self._model_name = self._engine_config.model
        self._model_artifact_key = model_artifact_key
        self._endpoint = endpoint
        self._artifacts = None

    @property
    def artifacts(self):
        if self._artifacts is None:
            raise ValueError("Artifacts not configured, run model.setup()")
        return self._artifacts

    @property
    def model_key(self):
        return self._model_artifact_key

    def _request_model(self, req_str):
        response = self._engine.oai_http_client.post(self._endpoint, content=req_str)
        status_code = response.status_code
        return json.dumps({"status": status_code, "data": response.text})

    def load_context(self, context):
        if self._engine is None:
            self._engine = self._engine_klass(config=self._engine_config)
        model_path = context.artifacts[self._model_artifact_key]
        self._engine.start_proc(model_path)

    def predict(self, context, model_input: List[str], params=None) -> List[str]:
        return [self._request_model(req) for req in model_input]

    def _hub_download_snapshot(self, repo_name: str, local_dir: str = "/root/models"):
        local_dir = local_dir.rstrip('/')
        model_local_path = f"{local_dir}/{repo_name}"
        snapshot_download(repo_id=self._model_name,
                          local_dir=model_local_path)
        return model_local_path

    def _setup_snapshot(self, local_dir: str = "/root/models"):
        return self._hub_download_snapshot(self._model_name, local_dir)

    def _setup_artifacts(self, local_dir: str = "/root/models"):
        local_path = self._setup_snapshot(local_dir)
        self._artifacts = {self._model_artifact_key: local_path}
        return self._artifacts

    def _verify_chat_template(self,
                              artifacts: Dict[str, str],
                              tokenizer_config_file: str = "tokenizer_config.json",
                              chat_template_key: str = "chat_template",
                              ):
        model_dir_path = Path(artifacts[self._model_artifact_key])
        tokenizer_config_file = model_dir_path / tokenizer_config_file
        assert tokenizer_config_file.exists(), f"Tokenizer config file not found at {str(tokenizer_config_file)}"
        with open(str(tokenizer_config_file), "r") as f:
            tokenizer_config = json.loads(f.read())
            chat_template = tokenizer_config.get(chat_template_key)
            if chat_template is None:
                raise ValueError(f"Chat template not found in tokenizer config file {str(tokenizer_config_file)}")

    def get_pip_reqs(self, **kwargs):
        return self._engine_config.default_pip_reqs(**kwargs)

    def setup(self,
              *,
              local_dir="/root/models",
              verify_chat_template: bool = True,
              tokenizer_config_file: str = "tokenizer_config.json",
              chat_template_key: str = "chat_template"):
        self._setup_artifacts(local_dir)
        if verify_chat_template is True:
            self._verify_chat_template(self.artifacts, tokenizer_config_file, chat_template_key)
        debug_msg(f"Command to be run: {self._engine_config.to_run_command()}")



import sys
from dataclasses import field, dataclass
from typing import List, Dict, Optional

from mlflow_extensions.serving.engines.base import EngineConfig, debug_msg, EngineProcess


@dataclass(frozen=True, kw_only=True)
class VLLMEngineConfig(EngineConfig):
    entrypoint_module: str = field(default="vllm.entrypoints.openai.api_server")
    enable_experimental_chunked_prefill: bool = field(default=False)
    max_num_batched_tokens: int = field(default=512)  # 512 is based on A100 ITL for llama model
    enable_prefix_caching: bool = field(default=False)
    vllm_command_flags: Dict[str, Optional[str]] = field(default_factory=dict)
    trust_remote_code: bool = field(default=False)
    max_model_len: Optional[int] = field(default=None)
    served_model_alias: Optional[str] = field(default=None)
    guided_decoding_backend: Optional[str] = field(default=None)

    def _to_vllm_command(self, local_model_path: Optional[str] = None) -> List[str]:
        flags = []
        skip_flags = ["--enable-chunked-prefill",
                      "--model",
                      "--max-num-batched-tokens",
                      "--enable-prefix-caching",
                      "--max-model-len",
                      "--trust-remote-code",
                      "--served-model-name",
                      "--guided-decoding-backend"]
        for k, v in self.vllm_command_flags.items():
            if k in skip_flags:
                debug_msg(f"Skipping flag {k} use the built in argument")
                continue
            flags.append(k)
            if v is not None:
                flags.append(v)
        if self.enable_experimental_chunked_prefill is True:
            flags.append("--enable-chunked-prefill")
            flags.append("--max-num-batched-tokens")
            flags.append(str(self.max_num_batched_tokens))
        if self.enable_prefix_caching is True:
            flags.append("--enable-prefix-caching")
        if self.trust_remote_code is True:
            flags.append("--trust-remote-code")
        if self.max_model_len is not None:
            flags.append("--max-model-len")
            flags.append(str(self.max_model_len))
        if self.guided_decoding_backend is not None:
            flags.append("--guided-decoding-backend")
            if self.guided_decoding_backend not in ["outlines", "lm-format-enforcer"]:
                raise ValueError(f"Invalid guided decoding backend {self.guided_decoding_backend}")
            flags.append(self.guided_decoding_backend)

        return [
            sys.executable,
            "-m",
            self.entrypoint_module,
            "--host",
            self.host,
            "--port",
            str(self.port),
            "--model",
            self.model if local_model_path is None else local_model_path,
            "--served-model-name",
            self.model if self.served_model_alias is None else self.served_model_alias,
            *flags,
        ]

    def to_run_command(self, local_model_path: Optional[str] = None) -> List[str]:
        return self._to_vllm_command(local_model_path=local_model_path)

    def engine_pip_reqs(self, *,
                        vllm_version: str = "0.5.5",
                        httpx_version: str = "0.27.0",
                        flash_attn_version: str = "2.6.3",
                        lm_format_enforcer_version: str = "0.10.6",
                        outlines_version: str = "0.0.46") -> List[str]:
        default_installs = [f"vllm=={vllm_version}", f"httpx=={httpx_version}", f"flash-attn=={flash_attn_version}"]
        if self.guided_decoding_backend == "lm-format-enforcer":
            default_installs.append(f"lm-format-enforcer=={lm_format_enforcer_version}")
        if self.guided_decoding_backend == "outlines":
            default_installs.append(f"outlines=={outlines_version}")
        return default_installs


class VLLMEngineProcess(EngineProcess):

    @property
    def engine_name(self) -> str:
        return "vllm-engine"

    def health_check(self) -> bool:
        try:
            resp = self.server_http_client.get("/health")
            return resp.status_code == 200
        except Exception as e:
            debug_msg(f"Health check failed with error {e}; server may not be up yet or crashed;")
            return False



import json
import typing
from typing import Optional

import httpx
from httpx import Client, Request, Response

if typing.TYPE_CHECKING:
    from openai import OpenAI
    from langchain_openai import ChatOpenAI


def is_local(url: str) -> bool:
    return "0.0.0.0" in url or "localhost" in url or "127.0.0.1" in url


def validate_url_token(url: str, token: Optional[str] = None) -> bool:
    if is_local(url) is False and token is None:
        return False
    return True


def build_endpoint_url(url: str) -> str:
    normalized_url = url.rstrip("/")
    if normalized_url.endswith("/invocations"):
        normalized_url = normalized_url.replace("/invocations", "")
    if is_local(normalized_url):
        return normalized_url.rstrip("/")
    return normalized_url


class CustomMLFlowHttpClient(Client):

    def __init__(self, *, endpoint_url: str,
                 token: Optional[str] = None,
                 model_name: Optional[str] = None,
                 timeout: int = 30):
        super().__init__()
        if validate_url_token(endpoint_url, token) is False:
            raise ValueError("You must provide a token unless the endpoint is localhost or 0.0.0.0")
        headers = {
            'Content-Type': 'application/json',
        }
        if token:
            headers['Authorization'] = f"Bearer {token}"
        base_url = build_endpoint_url(endpoint_url)
        self._http_client = httpx.Client(base_url=base_url, headers=headers, timeout=timeout)

    def send(self, request: Request, **kwargs) -> Response:
        content = request.content.decode("utf-8")
        content = {"inputs": [content]}
        response = self._http_client.post("/invocations", json=content)
        resp_data = json.loads(response.json()["predictions"][0])
        response_content = resp_data["data"]
        response_status = resp_data["status"]
        return Response(
            request=response.request,
            status_code=response_status,
            headers={"Content-Type": "application/json"},
            content=response_content)


def OpenAIWrapper(*args, **kwargs) -> 'OpenAI':
    kwargs.pop("http_client", None)
    assert "base_url" in kwargs, "You must provide a base_url"
    base_url = kwargs["base_url"]
    api_key = kwargs.get("api_key")
    if validate_url_token(base_url, api_key) is False:
        raise ValueError("You must provide an api_key")

    if "model" not in kwargs:
        raise ValueError("You must provide a model name, model=...")
    model = kwargs["model"]

    from openai import OpenAI
    return OpenAI(*args, **kwargs, http_client=CustomMLFlowHttpClient(
        endpoint_url=base_url,
        token=api_key,
        model_name=model
    ))


def ChatOpenAIWrapper(**kwargs) -> ChatOpenAI:
    kwargs.pop("http_client", None)
    assert "base_url" in kwargs, "You must provide a base_url"
    base_url = kwargs["base_url"]
    api_key = kwargs.get("api_key")
    if validate_url_token(base_url, api_key) is False:
        raise ValueError("You must provide an api_key")

    from langchain_openai import ChatOpenAI
    return ChatOpenAI(**kwargs, http_client=CustomMLFlowHttpClient(endpoint_url=base_url, token=api_key))



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="mlflow-extensions",
    author="Sri Tikkireddy",
    author_email="sri.tikkireddy@databricks.com",
    description="Extensions for mlflow to make the devloop better for custom models.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stikkireddy/mlflow-extensions",
    packages=find_packages(),
    install_requires=["httpx", "huggingface-hub", "filelock", "mlflow-skinny"],
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)
